; 
; This is a quick example of using the Extended Memory Manager
; (XMM) from a DOS program.
; 
      .model small,c
      .386
      .stack

      include support.asm
.data
      _xmm           dd 0
      xms_version    dw 0
      hEMB           dw 0

      ; Parameter block for the Move Extended Memory
      ; Block function.
      XMMMOVEPARS STRUCT
         blockLength dd 0
         hSrc        dw 0
         srcAddress  dd 0
         hDest       dw 0
         destAddress dd 0
      XMMMOVEPARS ENDS
      xmmmove XMMMOVEPARS <>
.code
.startup
; 
      ; Call the XMS v2+ Installation Check function to
      ; determine if the XMS driver (the XMM) is installed.
      mov   ax,4300h
      int   2fh
      cmp   al,80h
      je   @F      
      print NL,"XMM not installed",NL
      jmp   promptForExit
   @@:
      ; Call the XMS v2+ Get Driver Address function to get
      ; the entry point address of the XMS driver.
      mov   ax,4310h
      int   2fh
      mov   word ptr _xmm,bx
      mov   word ptr _xmm+2,es

      ; Call the Get XMS Version function.
      xor   ah,ah      
      call  _xmm
      mov   xms_version,ax
      
      ; And display it.
      push  ax
      shr   ax,8
      invoke w2dec,ax,ADDR outbuf
      print NL,"XMS version = ",ADDR outbuf
      pop   ax
      xor   ah,ah
      invoke w2dec,ax,ADDR outbuf
      print ".",ADDR outbuf,NL

      ; Call the Query Free Extended Memory function and
      ; display the return values. Use the 32-bit version
      ; for EMS v3+.
      .IF xms_version >= 300h
         mov   ah,88h
         call  _xmm
         invoke dw2dec,edx,ADDR outbuf
         print "Total Extended Memory = ",ADDR outbuf,"KB",NL
         invoke dw2dec,eax,ADDR outbuf
         print "Largest free block = ",ADDR outbuf,"KB",NL
      .ELSE
         mov   ah,8
         call  _xmm
         invoke dw2dec,dx,ADDR outbuf
         print "Total Extended Memory = ",ADDR outbuf,"KB",NL
         invoke dw2dec,ax,ADDR outbuf
         print "Largest free block = ",ADDR outbuf,"KB",NL
      .ENDIF

      ; Call the Allocate Extended Memory Block function
      ; and attempt to allocate a block twice the size of
      ; the largest free block. Use the 32-bit version for
      ; EMS v3+.      
      ;
      ; Under Windows 98 SE, which supports EMS v3.0, the
      ; largest free block seems to be stuck at 2048KB,
      ; even when a larger block can be allocated.   
      .IF xms_version >= 300h
         mov   edx,eax
         shl   edx,1
         mov   ah,89h
      .ELSE
         mov   dx,ax
         mov   ah,9
      .ENDIF
      call  _xmm
      test  ax,ax
      jnz   @F
      xor   bh,bh
      invoke dw2hex,bx,ADDR outbuf
      print "Allocation failed, error code = ",ADDR outbuf,"h",NL
      jmp   promptForExit
   @@:
      ; Save the handle.
      mov   hEMB,dx

      ; Call the Get EMB Handle Information function for the
      ; allocated block and display the returned values.
      mov   ah,0eh
      mov   dx,hEMB
      call  _xmm
      test  ax,ax
      jnz   @F
      xor   bh,bh
      invoke dw2hex,bx,ADDR outbuf
      print "Get EMB Handle Information failed, error code = ",ADDR outbuf,"h",NL
      jmp   freeBlock
   @@:      
      invoke w2dec,dx,ADDR outbuf
      print "Allocated block size = ",ADDR outbuf,"KB",NL
      push  bx
      xor   bh,bh      
      invoke w2dec,bx,ADDR outbuf
      print "Number of handles still available = ",ADDR outbuf,NL
      pop   bx
      shr   bx,8
      invoke w2dec,bx,ADDR outbuf
      print "Lock count (0 if not locked) = ",ADDR outbuf,NL      

      ; Set up and call the Move Extended Memory Block
      ; function to copy the display memory to extended
      ; memory.
      ;
      ; If either handle is 0, then the corresponding
      ; address is a far (segment:offset) address referring
      ; to directly addressable memory rather than a 32-bit
      ; offset address referring to the extended memory
      ; block.
      ;
      ; The function can copy memory between or within
      ; directly addressable memory and/or extended memory.
      ;
      ; The block length *must* be even.
      ;
      mov   xmmmove.blockLength,8000h
      mov   xmmmove.hSrc,0
      mov   xmmmove.srcAddress,0b8000000h
      mov   dx,hEMB
      mov   xmmmove.hDest,dx
      mov   xmmmove.destAddress,0
      mov   ah,0bh
      mov   si,OFFSET xmmmove
      call  _xmm
      test  ax,ax
      jnz   @F
      xor   bh,bh
      invoke dw2hex,bx,ADDR outbuf
      print "Move failed, error code = ",ADDR outbuf,"h",NL
      jmp   freeBlock
   @@:
      print NL,"Display memory copied to extended memory, "
      print "press any normal key to CLS...",NL
      call waitkey
      call cls
      print NL,"Press any normal key to copy display memory back...",NL
      call waitkey

      ; Under Windows 2000, and I assume under XP, CLS will
      ; not work when the app is running in a Window, and
      ; copying the contents of the display memory back will
      ; have no effect, either running in a window or full
      ; screen.
      mov   xmmmove.blockLength,8000h
      mov   dx,hEMB
      mov   xmmmove.hSrc,dx
      mov   xmmmove.srcAddress,0
      mov   xmmmove.hDest,0
      mov   xmmmove.destAddress,0b8000000h
      mov   ah,0bh
      mov   si,OFFSET xmmmove
      call  _xmm
      test  ax,ax
      jnz   freeBlock
      xor   bh,bh
      invoke dw2hex,bx,ADDR outbuf
      print "Move failed, error code = ",ADDR outbuf,"h",NL

   freeBlock:
      ; Call the Free Extended Memory Block function
      ; to free the allocated block.
      mov   ah,0ah
      mov   dx,hEMB
      call  _xmm
      test  ax,ax
      jnz   @F
      xor   bh,bh
      invoke dw2hex,bx,ADDR outbuf
      print "Free failed, error code = ",ADDR outbuf,"h",NL
   @@:     

   promptForExit:
      loc 6
      print NL,"Press any normal key to exit...",NL
      call waitkey
; 
.exit

end
